from TCPModemDriver import TCPModemDriver
from ModemDriver import ModemDriver
from SerialDriver import SerialDriver
from DeviceDriver import DeviceDriver
from ThreadedDriverWrapper import ThreadedDriverWrapper
from TCPTunnel import TCPTunnel
from UDPTunnel import UDPTunnel
from datetime import datetime
import logging
import serial
import os
import time
import threading

##########################
#codes
CODE_PLAYERINFO = b"&"
CODE_DIRECTORY = b"%"
CODE_DIAL = b"#"
CODE_ACKNOWLEDGE = b"$"
CODE_INCOMING = b"="
CODE_START = b"-"
CODE_CONNECT = b"*"
CODE_RESPONSE = b"*"
CODE_END = b"\xFF"
CODE_PULSE = b" "
CODE_CHECK_ACK = b"\\"
CODE_ACK = b"/"
CODE_ERROR = b"\xFE"
CODE_DIALONLY = b"\xFD"
CODE_FOUND = b"\xFC"
CODE_APPROVED = b"\xFA"

##########################
class AccessPoint():

    messageBuffer = b""

    
    threadedDriver : ThreadedDriverWrapper = None
    tcptunnel : TCPTunnel = None
    udptunnel : UDPTunnel = None

    logger = logging.getLogger('dreampi')
    state = "starting"

    def assignCommPort(self,com_port) -> str:
       
        if com_port.isdigit() and int(com_port) > 1000:
            deviceDriver = TCPModemDriver(int(com_port))
            self.threadedDriver = ThreadedDriverWrapper(deviceDriver)
            return "TCP"
            
        deviceDriver = None
        try:
            deviceDriver = ModemDriver(com_port,False)           
            if deviceDriver.serialDevice is not None:
                self.com_port = com_port
                
                return self.com_port
        except serial.SerialException as e:
            message = str(e)
            # print(message)
            #if "could not open port" in message:
            self.com_port = None
        except IOError as e:
            self.com_port = None
        finally:
            if deviceDriver is not None:
                deviceDriver.disconnect()
                deviceDriver = ModemDriver(com_port)
                self.threadedDriver = ThreadedDriverWrapper(deviceDriver)
        
        try:
            print("AP:Checking serial driver")
            deviceDriver = SerialDriver(com_port)
            if deviceDriver is not None and deviceDriver.serialDevice is not None:
                print("AP:Serial Name")
                print("AP:" + deviceDriver.serialDevice.name)
                self.com_port = com_port
                self.threadedDriver = ThreadedDriverWrapper(deviceDriver)
                return self.com_port
        except serial.SerialException as e:
            message = str(e)
            print("AP:Message %s" % message)
            if "could not open port" in message:
                self.com_port = None
        except IOError as e:
            self.com_port = None
         
        #assign to tcp
        self.com_port = None
        self.threadedDriver = None
        return self.com_port

    def scanCommPorts(self) -> str:
        for i in range(3,25): # this should be a big enough range. USB com ports usually end up in the teens.
            osName = os.name
            if osName == 'posix': # should work on linux and Mac for USB modem, but untested.
                com_port = "/dev/ttyACM%s" % i
            else:
                com_port = "COM%s" % i
                
            com_port = self.assignCommPort(com_port)
            
            if com_port:
                return com_port
                
        com_port = "TCP"
        deviceDriver = TCPModemDriver()
        self.threadedDriver = ThreadedDriverWrapper(deviceDriver)
        return com_port
            
    def sendDirectory(self):
        self.write(CODE_DIRECTORY)
        with open('directory.txt') as f:
            lines = f.readlines()
            for element in lines:
                self.write(element.encode())
            
        self.write(CODE_END)
        print("AP:Sent")
        
        
    def sendErrorMessage(self,message):
        self.write(CODE_ERROR)
        self.write(message.encode())
        self.write(CODE_END)
        pass
        
    def sendPlayerInfo(self):
        self.write(CODE_PLAYERINFO)
        print("AP:Openning playerinfo.txt")
        with open('playerinfo.txt') as f:
            lines = f.readlines()
            for element in lines:
                self.write(element.encode())
            
        self.write(CODE_END)
        print("AP:Sent")

    def getPlayerID(self) -> str:
        with open('playerinfo.txt') as f:
            lines = f.readlines()
            for element in lines:
                if "name:" in element:
                    return element[5:]
        return ""


    def disconnect(self):
        self.threadedDriver.disconnect()

    def readDevice(self):
        self.messageBuffer += self.threadedDriver.read()
       
        
    def readByte(self):
        self.readDevice()
        if len(self.messageBuffer) == 0:

            return b""
        
        b = self.messageBuffer[0:1]
        self.messageBuffer = self.messageBuffer[1:]
        #print("AP:readByte %s" % b)
        if b is None:
            b = b""
        return b
        
    def read(self):
        self.readDevice()
        if len(self.messageBuffer) == 0:
            return None
        
        b = self.messageBuffer[0:]
        self.messageBuffer = b""
        
        return b   
    
 
    def write(self,toSend):
        self.threadedDriver.write(toSend)
    

    def flush(self):
        self.threadedDriver.flush()
        
    def approved(self):
        print("AP:Sending ready")
        self.tcptunnel.ready()
        self.write(CODE_APPROVED)
        self.udptunnel = UDPTunnel()
        self.write(CODE_START)
        self.udptunnel.beginExchange(self,self.tcptunnel.remoteAddress)
        


    def dial(self):
        ipAddress = b""
        opponentName = b""
        char = self.readByte()
        
        while char != b'#':
            if len(char) > 0:
                ipAddress += char
            char = self.readByte()
        
        
        char = self.readByte()
        while len(char) > 0:
            if len(char) > 0:
                opponentName += char
                
            char = self.readByte()
        

        try:
            print("AP:Flushing buffer")
            self.flush()
        finally:
            pass
            
        print("AP:Dialing %s:%s" % (opponentName.decode() or "Opponent",ipAddress.decode()))
        
        print("AP:Initializing tunnel connection")
        self.state = self.tcptunnel.initConnection(ipAddress.decode(),self.getPlayerID(),opponentName.decode())
        
        
        
        if  "failed" in self.state:
            print("AP:"+self.state)
            self.sendErrorMessage(self.state)
            self.tcptunnel.disconnect()
            print("AP:Disconnecting from tunnel")
            return None
            
        self.remoteAddress = ipAddress.decode()
        print("AP:Initialized") 
        self.write(CODE_FOUND)
        
        
        
        
        
    def doNetworkRouting(self):
        while True:
           
            key,message = self.tcptunnel.getMessage()
            #print("AP:Got message")
            if not key:
                break;
          
            print("AP:Got message %s" % key)  
            if key == "incoming":
                self.write(CODE_INCOMING)
                self.write(message)
                continue
            if key == "listenEnabled" and message == "false": 
            
                self.write(CODE_DIALONLY)
                
                continue 
            if key == "readyip":
                
                
                return 1 
            if key == "acknowledge":
                print("AP:Approved")
                self.approved()
                return 1
                
          
            if not data:
                print("AP:failed to init")
    
        return 0
    
        #used when netlinkOS waits for acknowledge  
        #if char == b':':
            # print("Checking ack")
            # if self.tcptunnel.checkAcknowledge():
                # print("Approved")
                # self.approved()
            # while(self.waitThread):
                # pass 
            # return;
                    
                    
        #used when netlinkOS sends acknowledge  
            # 
                
                
    def doLocalRouting(self):
        self.tcptunnel = TCPTunnel()
        ipAddress = b""
        try:
            now = datetime.now()
            while (datetime.now() - now).total_seconds() < 360 and self.doNetworkRouting() == 0:
              
                char = self.readByte()
                
                if not char:
                    continue
                    
                print("AP:Routing: %s" % char)
                now = datetime.now() 
                
                #used when netlinkOS requests player info
                if char == CODE_PLAYERINFO:
                    self.sendPlayerInfo()
                if char == CODE_CONNECT:
                    return False
                #used when netlinkOS requests directory
                if char == CODE_DIRECTORY:
                    self.sendDirectory()
                    
                #used when netlinkOS is dialing   
                if char == CODE_DIAL:
                    print("AP:Dial command")
                    self.dial()
                
                if char == CODE_ERROR:
                    pass
                    
                if char == CODE_ACKNOWLEDGE:
                    print("AP:Sending acknowledge");
                    self.tcptunnel.sendAcknowledge()
            
            if((datetime.now() - now).total_seconds() >= 360):
                print("AP:Timed out")

            else:
                self.tcptunnel.disconnect()

               
                
               
        except IOError as error:
            print("AP:"+str(error))
            #ex_type, ex_value, ex_traceback = sys.exc_info()
            
            #print("AP:Exception type : %s" % ex_type.__name__)
            #print("AP:Exception message : %s" %ex_value)
            self.disconnect()
            return False
        except Exception as error:
            
            print("AP: Error "+str(error))
            self.disconnect()
            return False
        finally:
            print("AP:Done");
            
        self.flush()
        return True
        
    def doTunnelRouting(self):
        self.udptunnel = UDPTunnel()          
        self.write(CODE_START)
        self.udptunnel.beginExchange(self.threadedDriver,self.tcptunnel.remoteAddress,True)
        

    def connect(self):
        print("AP:Attempting to connect")
       
        isConnected = self.threadedDriver.connect()
        
        if not isConnected:
            print("AP:Netlink Error")
            return
            
        print("AP:Sending response")
        self.write(CODE_RESPONSE);
        
        print("AP:AccessPoint connected to device")
        
        time_digit_heard = None

        print("MN:Access point routing")
        if self.doLocalRouting():
            print("MN:Begin Tunnel Routing")
            self.doTunnelRouting()
            
        self.disconnect()
        return
        
    def disconnect(self):
        print("AP:Disconnecting")
        if self.tcptunnel:
            self.tcptunnel.disconnect()
        if self.threadedDriver:
            self.threadedDriver.disconnect()



    def __init__(self,port = None):  
        if port is None:
            com_port = self.scanCommPorts();
        else:
            com_port = self.assignCommPort(port);
    
        if com_port:
            print("AP:Device found at %s" % com_port)
        else:
            print("AP:No device found")
            sys.exit()