# -*- coding: utf-8 -*-
"""
Created on Thu May 19 08:01:31 2022

@author: joe
"""
#netlink_version=1663872260.57
import sys

if __name__ == "__main__":
    print("TT:This script should not be run on its own")
    sys.exit()

from datetime import datetime

import socket
import time
import logging
import threading
import binascii
import select



class TCPTunnel():
    logger = logging.getLogger('dreampi')
    
    tcpThread:threading.Thread
   
    printout = False
    isAlive = True
    listenEnabled = True
    if 'printout' in sys.argv:
        printout = True
    
    remoteAddress = ""

    tcpListener : socket.socket = None
    tcpClient : socket.socket = None
    tcpCurrent : socket.socket = None
    
    msgBuffer = b""
    messages = []
    

    def __init__(self):  
        
        print("TT: Thread started")
        self.tcpThread = threading.Thread(target=self.listeningThread, daemon=True)
        self.tcpThread.start()
    
    
    def initConnection(self,dial_string,playerID,opponentName):
        print("TT:Initializing for:%s" % dial_string)
        
        ip_set = dial_string.replace('*','.').split('.')
        for i,set in enumerate(ip_set): #socket connect doesn't like leading zeroes now
            fixed = str(int(set))
            ip_set[i] = fixed
            
        self.remoteAddress = ('.').join(ip_set)

        print("TT:I'm calling")
        PORT = 65432
        
        
        self.tcpListener.close()
        self.tcpClient = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.tcpClient.settimeout(120)
        
        try:
            print("TT:Attempting to connect to %s" % opponentName.strip() or "Opponent")
            
            self.tcpClient.connect((self.remoteAddress, PORT))
            
            
            self.tcpCurrent = self.tcpClient
            self.tcpThread = threading.Thread(target=self.messageHandler, daemon=True)
            self.tcpThread.start()
            
            print("TT:Connection successful")
            self.sendIncoming(playerID)
            return "success"
                
        except Exception as error:
            print("TT:"+str(error))
            self.remoteAddress = ""
            strError = str(error)
            strMessage = f'failed: {strError}' 
           
            return strMessage



    
    def listeningThread(self):
        print("TT:Attempting to listen")
        try:
            
            PORT = 65432
            self.tcpListener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.tcpListener.setblocking(True)
            self.tcpListener.bind(('', PORT))
            self.tcpListener.listen(1)
            print("TT:I'm listening on port: %d" % PORT)
            
        except:
            print("TT:Port %d in use, dial allowed only" % PORT)
            self.listenEnabled = False
            self.messages.append(('listenEnabled', "false"))
            return
        
        
        print("TT:Waiting on accept")
        if self.accept() == False:
            print("TT:Rejected")
            return
        
                
        print("TT:Accepted")
        
        self.messageHandler()
        print("TT: Thread stopped")

    def messageHandler(self): 
        self.tcpCurrent.settimeout(0.01)
        while self.isAlive:
            try:
                self.messageProcessor()
            except socket.error: #first try can return no payload
                continue
            
     
        
        
    def accept(self) -> bool:
    
        
        try:
            ready = select.select([self.tcpListener], [], [])
            if ready[0]:
                print("TT:Accepting")
                self.tcpCurrent, addr = self.tcpListener.accept()
                
                self.remoteAddress = addr[0]
                return True
        except:
            return False        


        

    def messageProcessor(self):
       # print("TT: Process %s %d" % (self.msgBuffer,len(self.msgBuffer.split(b'|'))))
        if len(self.msgBuffer.split(b'|')) <= 1: 
            self.msgBuffer += self.tcpCurrent.recv(1024)
            
            
        if len(self.msgBuffer) == 0:
            return
            
        print("TT: Buffer %s" % self.msgBuffer)
        packets = self.msgBuffer.split(b'|')
        
        for packet in packets:
            if not b":" in packet:
                break
            breakdown = packet.split(b":")
            self.messages.append((breakdown[0].decode(), breakdown[1]))
            
            self.msgBuffer = self.msgBuffer[len(packet) + 1:] # plus one for the pipe
        
        return
 
    def disconnect(self):
        self.isAlive = False
        if self.tcpClient:
            self.tcpClient.close()
        
        if self.tcpListener:
            self.tcpListener.close()
        
        
            
            

    
    def getMessage(self):
        if len(self.messages) == 0:
            return (None,None)
        message = self.messages[0]
        self.messages = self.messages[1:]
        return message

    
    # ####################################
    # #Client send commands
    # def checkAcknowledge(self) -> bool:
        # print("TT:blocking")
        # ready = select.select([self.tcpClient], [], [])
        # print("TT:unblocking")
        
        # if ready[0]:
            # data = self.messageProcessor()
            # print("TT:TCP ack %s" % data)
            # if data == b'ack':
                # return True
        # return False       
    
    def ready(self):
        try:
            
            self.tcpCurrent.sendall(b"readyip:|")
            print("TT:Ready sent")
            # ready = select.select([self.tcpCurrent], [], [])
            # if ready[0]:
                # data = self.messageProcessor()
                # print("TT:TCP %s" % data)
                # if data == b'g2gip':
                    # #notify device a tunnel is established
                   
                    # print("TT:Ready for Data Exchange!")
                    # #tcp.shutdown(socket.SHUT_RDWR)
                    # #tcp.close()
                    # return "connected"
                # return "failed"
        except socket.error:
            print("TT:"+str(socket.error))
            self.remoteAddress = ""
            return "failed"
            
    def sendIncoming(self,playerID):
        print("TT:Sending incoming")
        self.tcpCurrent.sendall(b"incoming:%s|" % playerID.strip().encode())
            
        
    ####################################
    #Listener send commands
    def begin(self):
        self.tcpCurrent.sendall(b'g2gip:|')
    
    def sendAcknowledge(self):
        self.tcpCurrent.sendall(b'acknowledge:|')
        