import time
import socket
import select
import os
import sys
import atexit
import threading

from datetime import datetime
from DeviceDriver import DeviceDriver
class ThreadedDriverWrapper:

    readTime = 0.0
    deviceDriver: DeviceDriver = None
    writeData: bytearray = b""
    readData: bytearray = b""
    timeout = 0.1
    
    peername = ""
    isAlive = False
    writelock = threading.Lock()
    readlock = threading.Lock()
    
    readThread:threading.Thread
    writeThread:threading.Thread
    
    
    def __init__(self,deviceDriver):
        self.deviceDriver = deviceDriver
        atexit.register(self.exit_handler)
        
    def exit_handler(self):
        pass    

    def connect(self) -> bool:
        if self.isAlive:
            return True
        
        print("TW:Connecting")
        self.isAlive = self.deviceDriver.connect()
        
        
        self.readThread = threading.Thread(target=self.reading, daemon=True)
        self.writeThread = threading.Thread(target=self.writing, daemon=True)
        
        self.readThread.start()
        self.writeThread.start()
        
        return self.isAlive          
    
     
       
    def reading(self):

        print("TW:Read thread started")
        
        while self.isAlive:
            try:
                
                if not self.deviceDriver:  
                    continue
               
                data = self.deviceDriver.read()
                
                try:
                    self.readlock.acquire()
                    self.readData += data
                    

                except:
                    raise
                finally:
                    self.readlock.release()
                    
            except Exception as ex:
                if "close" in str(ex):
                    print("TW:Error %s" % str(ex))
                    self.disconnect()
                    break
                if "aborted" in str(ex):
                    print("TW:Error %s" % str(ex))
                    self.disconnect()
                    break

        print("TW:Read thread ended")
        
    

    
    def writing(self):
        print("TW:Write thread started")
        while self.isAlive:
            try:
                
                if not self.deviceDriver or len(self.writeData) <= 0:
                    continue

                try:
                    self.writelock.acquire()
                    self.deviceDriver.write(self.writeData)
                    self.writeData = b''
                except:
                    raise
                finally:
                    self.writelock.release()

            except Exception as ex:
                if "close" in str(ex):
                    print("TW:Error %s" % str(ex))
                    self.disconnect()
                    break
                if "aborted" in str(ex):
                    print("TW:Error %s" % str(ex))
                    self.disconnect()
                    break
                print("TW:Error %s" % str(ex))
            
                
                
        print("TW:Write thread ended")
        
    def read(self ) -> bytearray:
        result = b""
        try:
            self.readlock.acquire()
            if self.isAlive == False:
                raise Exception("TW: Error: Device disconneted")
            
            
            if self.readData and len(self.readData) > 0:
                result = self.readData
                self.readData = b""
            
            
        except Exception as ex:
            print(ex)
            raise
        finally:    
            self.readlock.release()
        return result
        
        
        
    def readByte(self ) -> bytearray:
        result = b""
        try:
            self.readlock.acquire()
            
            
            if self.readData and len(self.readData) > 0:
                result = self.readData[0:1]
                self.readData = self.readData[1:]
            
            print("TW:readByte %s" % result)
        except Exception as ex:
            print(ex)
        finally:
            self.readlock.release()
        return result
         
        
    def write(self,data):      
        self.writelock.acquire()
        
        self.writeData += data
        self.writelock.release()

    

    def flush(self):
        self.deviceDriver.flush()
        self.readData = b""
        
    def disconnect(self):
        self.isAlive = False
        if self.deviceDriver:
            self.deviceDriver.disconnect()
            
            
